(* ==============================================================
// Copyright (c) 1996 by Aaron Castro and VK Thakur. 
// Version 3.0 enhancements by Eric Engler.
*)

{ Low-level string searching routines.  }
unit Sercheng;
{$X+}
interface
uses
  SysUtils;

function CompareStrings(Target, Pattern: PChar;
            CaseSensitive: Boolean; UseWildCards: Boolean): boolean;
function IsPatternWild(pattern: PChar): Boolean;
procedure SetMultiPat(MultiPat: PChar);
procedure SetSinglePat(SinglePat: Char);
function GetMultiPat: PChar;
function GetSinglePat: Char;

implementation

const
   MAXMULTIPAT = 5;
var
   pMultiPat: array[0..MAXMULTIPAT] of Char;
   pSinglePat: Char;
   MultWCsiz : integer;

{ True is returned if a wildcard appears anywhere in pattern,
  which is NULL terminated. }
function IsPatternWild(pattern: PChar): Boolean;
begin
  Result := StrPos(pattern, pMultiPat) <> nil;
  if not Result then
     Result := StrScan(pattern, pSinglePat) <> nil;
end;

{ Important! This is a recursive function.  Be wary of making
  off-the-cuff changes here. }
function CompareStrings(Target, Pattern: PChar;
            CaseSensitive: Boolean; UseWildCards: Boolean): Boolean;

   { Removes non printing characters at the end of a string }
   { This function is local to function CompareStrings }
   function StripNPAtTail(StrPtr: PChar): PChar;
   var
      P: PChar;
   begin
      P := StrPtr + StrLen(StrPtr);
      while Ord(P^) < 32 do  { while a non-printing control character }
      begin
         P^ := #0;   { end the string here }
         P := P - 1; { decrement pointer }
      end;
      Result := StrPtr;
   end;

   { This function is local to function CompareStrings. }
   { Find substring "Pattern" within string "Target", allowing for
     the possible use of wild cards. }
   function WildMatch(Target, Pattern: PChar): Boolean;
   var
      NextWC, P: PChar;
      SearchStr: Array[0..255] of char;
   begin
      { StrComp result = 0 if strings are same len and have same data }
      { StrLCmp is like StrComp, but only checks up to MAXLEN char's }
      { NOTE: Pattern is user's search key }

      { Look for exact match (incl len) bet. search key and
        pMultiPat wildcard }
      if 0 = StrComp(Pattern, pMultiPat) then
         Result := True  { Pattern consists on only a pMultiPat;
                           we'll match everything! }
      else if (Target^ = Chr(0)) and (Pattern^ <> Chr(0)) then
         Result := False { no more data in target string, but
                           still have data in pattern to locate;
                           this constitues a mismatch }
      else if Target^ = Chr(0) then
         Result := True  { no more data in target string or pattern;
                           this is a match }
        { Does the key Pattern start with a pMultiPat? }
      else if 0 = StrLComp(Pattern, pMultiPat, MultWCsiz) then
         begin
            { Key Pattern begins with pMultiPat, but may have more data
              after that. Check to see if we have ANY more wildcards
              in the rest of the search key: }
            if (not IsPatternWild(Pattern + MultWCsiz)) then
            begin
               { We do NOT have any more wild cards after this one. }
               { See if the rest of the search key can be found in
                 the current target string.  If it is found, we will use
                 P to skip over any no. of leading chars in the target -
                 which is what we need to do bec. of the pMultiPat. }
               P := StrPos(Target, Pattern + MultWCsiz);
               if (P <> nil)
                 { Yes, somewhere in the target is the rest of the key. }
                 { Did the remainder of the search key exactly match the
                   remainder of the target string - incl the same length ? }
                and (((StrLen(P) = StrLen(Pattern + MultWCsiz))
                  { Will they match exactly if we filter end-of-line junk? }
                or (StrLen(StripNPAtTail(P))=StrLen(Pattern+MultWCsiz)))) then
                  { We exactly matched the remainder of the key string after
                    we removed the initial pMultiPat (and the corresponding
                    arbitrary no. of chars at the beg. of target) from
                    consideration }
                  Result := True
               else
                  Result := False;
            end { end "if not IsPatternWild..." }
            else
            begin
               { We had a leading pMultiPat in our key, and have more wildcards }
               { Find out which wild card comes next }
               NextWC := StrPos(Pattern + MultWCsiz, pMultiPat);
               P := StrScan(Pattern + MultWCsiz, pSinglePat);
               { Determine which match came first, and save it in NextWC }
               if (NextWC = nil) then
                  NextWC := P
               else if (P <> nil) and (P < NextWC) then
                  NextWC := P;

               { SearchStr is used as a temporary buffer. }
               { Slide the key pattern to left; erasing the pMultiPat. }
               { End this new string just before the next wildcard. }
               StrLCopy(SearchStr,Pattern+MultWCsiz,NextWC-Pattern-MultWCsiz);

               { Does the key text after the pMultiPat exist in target? }
               P := StrPos(Target, SearchStr);

               { We didn't find the key text located after the pMultiPat, or
                 we found it at the very start!}
               { *** EWE bug fix: it's OK if P=Target 
                 OLD   if (P = nil) or (P = Target) then **** }
               if (P = nil) then  { new code for if stmt }
                  Result := False
               else
                  { We are matching the key text after the pMultiPat, so far ..}
                  Result := WildMatch(P + StrLen(SearchStr), NextWC);
            end;
         { We're ignoring case of a pMultiPat that isn't at beg. of string }
         end
      else if Pattern^ = pSinglePat then
         begin
            { skip over 1 char for each single-char wild card in key }
            repeat
               Inc(Pattern);
               Inc(Target);
            until (Pattern^ <> pSinglePat);
            { Found a non-wildcard char: make a recursive call }
            Result := WildMatch(Target, Pattern);
         end
      else if Target^ = Pattern^ then
         begin
            { We matched key text; skip over each set of matches }
            repeat
               Inc(Pattern);
               Inc(Target);
            until (Pattern^ = Chr(0)) or (Target^ = Char(0))
             or (Target^ <> Pattern^);
            { Done matching some text, now make recursive call }
            Result := WildMatch(Target, Pattern)
         end
      else
         Result := False;  { Oh well, ... }
   end;

{ Beginning of meat for function CompareStrings }
begin
   MultWCsiz := StrLen(pMultiPat);

   { if not case sensitive then make strings uppercase }
   if (not CaseSensitive) then
   begin
      StrUpper(Target);
      StrUpper(Pattern);
   end;
   if UseWildCards then
   begin
      Result := WildMatch(Target, Pattern);
   end
   else
      { This is MUCH faster! }
      Result := (StrPos(Target, Pattern) <> nil);
end;

procedure SetMultiPat(MultiPat: PChar);
begin
   StrLCopy(pMultiPat, MultiPat, MAXMULTIPAT);
end;

procedure SetSinglePat(SinglePat: Char);
begin
   pSinglePat := SinglePat;
end;

function GetMultiPat: PChar;
begin
   Result := pMultiPat;
end;

function GetSinglePat: Char;
begin
   Result := pSinglePat;
end;

begin
   { These are the standard Paradox wildcards: }
   StrPCopy(pMultiPat, '..');    { for SQL std: '%' }
   pSinglePat := '@';            { for SQL std: '_' }
end.
